/*****************************************************
Simple firmware for stepping a stepper motor at one step
per a constant time interval. The step interval can be
changed by redefining COMPARE_VALUE.


LED Blinking: Tracking engaged
LED Solid: Standby

Single press in standby -> Start tracking
Single press while tracking -> Pause
Long Press in any state -> Reverse motor direction as long as button is held. Return to standby state.

Motor Dir: PC0
Motor Step: PC1
Reset: PC2
LED: PC3

Outputs on PORTC.

AVR Type:  ATmega168A
AVR Clock: 7.3728MHz Crystal
*****************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define TRACK_SPEED 8036
#define REVERSE_SPEED 100
#define BLINK_LEVEL 25

#define LOOP_DELAY 20
#define LONG_PRESS_LEN (400/LOOP_DELAY)
#define SHORT_PRESS_LEN (80/LOOP_DELAY)

#define NO_PRESS 0
#define SHORT_PRESS 1
#define LONG_PRESS 2

uint8_t gotoIdleState(void);
uint8_t gotoTrackingState(void);
uint8_t gotoReverseState(void);
uint8_t buttonPressType(uint16_t buttonReg);

ISR(TIMER1_COMPA_vect) {
	// Clock the stepper motor driver
	PORTC |= (1<<PC1);
	_delay_us(10);
	PORTC &= ~(1<<PC1);
}

// Main
int main(void) {
	uint8_t currentState, led_blink_counter, buttonState;
	uint16_t buttonReg;

	// Init I/O
	DDRC = (1<<PC0)|(1<<PC1)|(0<<PC2)|(1<<PC3)|(0<<PC4)|(0<<PC5);
	PORTC = (1<<PC2)|(1<<PC3); // First turn off LED, and enable pull-up for the button
	
	// Initialize timer 1 constant interrupts, used to set timing for servo update signals
	TCCR1B = (1<<WGM12)|(0<<CS12)|(1<<CS11)|(1<<CS10);	// CTC mode, 1:64 prescaler
	
	buttonReg = 0;
	led_blink_counter = 0;
	currentState = gotoIdleState();
	sei();

	while (1) {
	
		// Keep track of button presses
		if (PINC & (1<<PC2)) {
			// User has released the button
			if (buttonReg > SHORT_PRESS_LEN) {
				buttonState = SHORT_PRESS;
			}else {
				buttonState = NO_PRESS;
			}
			buttonReg = 0;
		}else {
			// User is holding down the button
			if (buttonReg > LONG_PRESS_LEN) {
				buttonState = LONG_PRESS;
			}else {
				buttonState = NO_PRESS;
				buttonReg++;
			}
		}
		
		// Operate the state machine
		switch(currentState) {
		
			// Default standby state
			case 0: {
				// If short press goto state 1
				// else if long press goto state 2
				if (buttonState == SHORT_PRESS) {
					currentState = gotoTrackingState();
				}else if (buttonState == LONG_PRESS) {
					currentState = gotoReverseState();
				}
				
			}break;
			
			// Tracking state
			case 1: {
				// Blink the LED
				if (led_blink_counter > BLINK_LEVEL) {
					PORTC &= ~(1<<PC3);	// LED on
					led_blink_counter = 0;
				}else{
					PORTC |= (1<<PC3);		// LED off
					led_blink_counter++;
				}
				
				// If short press goto state 0
				// if long press goto state 2
				if (buttonState == SHORT_PRESS) {
					currentState = gotoIdleState();
				}else if (buttonState == LONG_PRESS) {
					currentState = gotoReverseState();
				}
			
			}break;
			
			// Reversing the motor state
			case 2: {
				// if no longer a long press, goto state 0
				if (buttonState != LONG_PRESS) {
					currentState = gotoIdleState();
				}
			
			}break;
			
			default: {
				currentState = 0;
			}
		}
		
		// Slight delay used to blink LED, check button status, and save polling
		_delay_ms(LOOP_DELAY);
	}
	
	return 0;
}

// Set state machine to intial state
uint8_t gotoIdleState() {
	PORTC &= ~(1<<PC3);	// LED on
	TIMSK1 = 0;				// Timer off
	PORTC |= (1<<PC0);		// Set step direction
	return 0;
}

// Set state machine to tracking state
uint8_t gotoTrackingState() {
	OCR1A = TRACK_SPEED;	// Slow step rate
	TIMSK1 = (1<<OCIE1A);	// Timer on
	PORTC |= (1<<PC0);		// Set step direction
	return 1;
}

// Set state machine to reversing state
uint8_t gotoReverseState() {
	OCR1A = REVERSE_SPEED;	// Fast step rate
	PORTC |= (1<<PC3);	// LED off
	TIMSK1 = (1<<OCIE1A);	// Timer on
	PORTC &= ~(1<<PC0);	// Set step direction
	return 2;
}

